'''
All about kernels

Also doctest other parts of this sub-module:
    >>> import doctest
    >>> doctest.testmod(definitions)
    TestResults(failed=0, attempted=5)
    >>> doctest.testmod(integral)
    TestResults(failed=0, attempted=9)
'''

from .definitions import *
from .integral import *

